#!/usr/bin/python
# $Id: $

DEBUG = True
ENABLE_SOUND = True

import embryo
import math
import random
import re
import sys
import traceback

import pyglet
from pyglet.gl import *
from pyglet.window import key, mouse

pyglet.resource.path.append('res')
pyglet.resource.reindex()
pyglet.resource.add_font('imperator.ttf')

class Prop(object):
    def __init__(self, filename, mask_filename, x, y, z, scale):
        self.image = pyglet.resource.texture(filename)
        glTexParameteri(self.image.target, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(self.image.target, GL_TEXTURE_WRAP_T, GL_CLAMP)

        if mask_filename == filename:
            self.mask_image = self.image
        else:
            self.mask_image = pyglet.resource.texture(mask_filename)

        self.filename = filename
        self.mask_filename = mask_filename
        self.x = x
        self.y = y
        self.z = z
        self.scale = scale
        self.opacity = 255
        self.recreate()

        self.hidden = False

        self.save_x = self.x
        self.save_y = self.y
        self.save_z = self.z
        self.save_scale = self.scale

    def recreate(self):
        w = self.image.width * self.scale
        h = self.image.height * self.scale
        x1 = self.x - w / 2
        x2 = self.x + w / 2
        y1 = self.y
        y2 = self.y + h
        z = self.z

        self.vertices = pyglet.graphics.vertex_list(4,
            ('v3f', (x1, y1, z,
                     x2, y1, z,
                     x2, y2, z,
                     x1, y2, z)),
            ('t3f', self.image.tex_coords),
            ('c4B', (self.opacity,) * 16))

    def draw(self):
        if self.hidden:
            return

        glBindTexture(self.image.target, self.image.id)
        glEnable(self.image.target)
        self.vertices.draw(GL_QUADS)

    def draw_hit_test(self):
        if self.hidden:
            return

        glBindTexture(self.mask_image.target, self.mask_image.id)
        glEnable(self.mask_image.target)
        self.vertices.draw(GL_QUADS)

class Camera(object):
    def __init__(self, x, y, z, yaw):
        self.x = x
        self.y = y
        self.z = z
        self.yaw = yaw

    def set_from(self, other):
        self.x = other.x
        self.y = other.y
        self.z = other.z
        self.yaw = other.yaw

    def clone(self):
        return Camera(self.x, self.y, self.z, self.yaw)

    def __repr__(self):
        return 'Camera(%f, %f, %f, %f)' % (self.x, self.y, self.z, self.yaw)

class CameraMotion(object):
    def __init__(self, camera, src, dest, time=1.0, callback=None):
        self.camera = camera
        self.src = src
        self.dest = dest
        self.max_time = time
        self.time = time
        self.camera.set_from(src)
        self.callback = callback
        pyglet.clock.schedule(self.update)

    def update(self, dt):
        if dt > 0.1:
            dt = 0 # work around pyglet oddity (er, bug).

        self.time -= dt
        if self.time <= 0:
            self.camera.set_from(self.dest)
            pyglet.clock.unschedule(self.update)
            if self.callback:
                self.callback()
            return

        f = self.time / self.max_time
        f = math.sin(f * math.pi / 2) ** 2
        self.camera.x = f * self.src.x + (1 - f) * self.dest.x
        self.camera.y = f * self.src.y + (1 - f) * self.dest.y
        self.camera.z = f * self.src.z + (1 - f) * self.dest.z
        self.camera.yaw = f * self.src.yaw + (1 - f) * self.dest.yaw

class Game(pyglet.window.Window):
    def __init__(self, *args, **kwargs):
        super(Game, self).__init__(*args, **kwargs)

        self.cutscene_labels = [
            pyglet.text.Label("Loading...",
                              font_name='Imperator',
                              font_size=13,
                              anchor_x='right',
                              anchor_y='bottom',
                              x=590, y=10,
                              color=(255,255,255,255)),
        ]
        self.loading = True

        self.bg_color = 0
        self.mouse_press = None
        self.fade = 0
        self.init_gl()

    def load_all(self, dt):
        self.load()
        self.init_bar()
        self.init_sound()

        self.camera = Camera(0, 0, 0, 0)
        self.set_camera(0, True)
        self.fade = 0
        self.BG_COLOR = 0.19
        self.bg_color = self.BG_COLOR

        self.init_gameplay()
        self.play(self.sound_intro)

        self.card = self.card_start
        self.card_y = 0
        self.cutscene = False
        self.cutscene_labels = []

        if DEBUG:
            self.push_handlers(GameDebug(self))
            self.CHEAT_SUCCESS = False

        self.loading = False
        self.invalid = True
        pyglet.clock.unschedule(self.load_all)

    # Gameplay
    # --------

    def init_gameplay(self):
        for prop in self.props:
            prop.hidden = False
            prop.opacity = 255

        self.get_prop('door_o').hidden = True
        self.get_prop('keys').hidden = True
        self.get_prop('friend').hidden = True
        self.get_prop('friend_jacket').hidden = True
        self.get_prop('phone_up').hidden = True

        self.have_phoned_mum = False
        self.have_boyfriend_offered_keys = False
        self.have_keys = False
        self.have_unlocked_cupboard = False
        self.have_boyfriend = True
        self.have_phonenumber = False
        self.have_phoned_friend = False
        self.have_friend = False
        self.have_warm_friend = False

        self.hand = None
        self.hand_y = 0
        self.panic_time = 0

        for prop in self.props:
            prop.x = prop.save_x
            prop.y = prop.save_y
            prop.z = prop.save_z
            prop.scale = prop.save_scale
            prop.recreate()
        self.resort_props()

    def select_outside(self):
        if self.hand and self.hand is not self.hand_kite:
            self.select_prop(self.get_prop('grass'))
        else:
            if self.CHEAT_SUCCESS:
                self.go_outside()
            elif self.have_boyfriend and not self.have_keys:
                self.set_bar_message(self.speaker_boyfriend,
                    "Babe, you know you can't go outside by yourself.")
                self.play(self.sound_bf_outside)
            else:
                self.go_outside()

    def select_prop(self, prop):
        if self.hand is self.hand_pills:
            if prop.filename == 'kitchen.png':
                self.flush_pills()
                self.hand = None
                self.play(self.sound_flush)
            else:
                self.get_prop('pills').hidden = False
                self.hand = None
                self.play(self.sound_putdown)
        elif self.hand is self.hand_phone:
            if prop.filename == 'photo.png':
                self.call_mum()
            elif (prop.filename == 'bookshelf.png' and self.have_phonenumber
                  and not self.have_phoned_friend):
                self.call_friend()
            else:
                self.hand = None
                self.get_prop('phone_up').hidden = True
                self.get_prop('phone_down').hidden = False
                self.play(self.sound_putdown)
        elif self.hand is self.hand_keys:
            self.hand = None
            if prop.filename == 'door_c.png':
                self.unlock_cupboard()
                self.play(self.sound_unlock)
            else:
                self.get_prop('keys').hidden = False
                self.play(self.sound_putdown)
        elif self.hand is self.hand_kite:
            self.hand = None
            self.get_prop('kite').hidden = False
            self.play(self.sound_putdown)
        elif self.hand is self.hand_jacket:
            self.hand = None
            if prop.filename == 'friend.png':
                self.get_prop('friend').hidden = True
                self.get_prop('friend_jacket').hidden = False
                self.have_warm_friend = True
                self.play(self.sound_jacket)
            else:
                self.get_prop('jacket').hidden = False
                self.play(self.sound_putdown)
        elif self.hand:
            self.hand = None
            self.play(self.sound_putdown)
        else:
            if prop.filename == 'bookshelf.png' and not self.have_boyfriend:
                self.hand = self.hand_yearbook
                self.have_phonenumber = True
                self.play(self.sound_pickup)
            elif prop.filename == 'door_c.png':
                if self.have_unlocked_cupboard:
                    prop.hidden = True
                    self.get_prop('door_o').hidden = False
                else:
                    self.rattle_cupboard()
                    self.play(self.sound_bf_cupboard)
            elif prop.filename == 'door_o.png':
                prop.hidden = True
                self.get_prop('door_c').hidden = False
            elif prop.filename == 'pills.png':
                prop.hidden = True
                self.hand = self.hand_pills
                self.play(self.sound_pickup)
            elif prop.filename == 'phone_down.png':
                self.hand = self.hand_phone
                prop.hidden = True
                self.get_prop('phone_up').hidden = False
                self.play(self.sound_pickup)
            elif prop.filename == 'keys.png':
                prop.hidden = True
                self.hand = self.hand_keys
                self.play(self.sound_pickup)
            elif prop.filename == 'jacket.png':
                prop.hidden = True
                self.hand = self.hand_jacket
                self.play(self.sound_pickup)
            elif prop.filename == 'kite.png':
                prop.hidden = True
                self.hand = self.hand_kite
                self.play(self.sound_pickup)
            elif prop.filename == 'kitchen.png':
                if self.have_boyfriend:
                    self.set_bar_message(self.speaker_boyfriend,
                        """Don't look at me, doing the dishes isn't my job.
                        """)
                    self.play(self.sound_bf_sink)
                elif self.have_friend:
                    self.set_bar_message(self.speaker_friend,
                        """I think you totally did the right thing.  He
                        put you through hell, and for what?
                        """)
                    self.play(self.sound_friend_sink)
            elif prop.filename == 'photo.png':
                if self.have_boyfriend and not self.have_phoned_mum:
                    self.set_bar_message(self.speaker_boyfriend,
                        """Your mum and I have been trying for years to help
                        you.  What more do you want us to do?
                        """)
                    self.play(self.sound_bf_mum)
                elif self.have_friend:
                    self.set_bar_message(self.speaker_friend,
                        """Your mum really wasn't here for you, was she?
                        Don't worry, we'll work things out.""")
                    self.play(self.sound_friend_mum)
            elif prop.filename == 'boyfriend.png':
                if not self.have_keys:
                    self.set_bar_message(self.speaker_boyfriend,
                        """What's wrong with you, anyway?
                        """)
                    self.play(self.sound_bf_bf)
                else:
                    self.set_bar_message(self.speaker_boyfriend,
                        """Let's get going.  Your mum's waiting.
                        """)
                    self.play(self.sound_bf_bf)
            elif prop.filename == 'friend.png':
                self.set_bar_message(self.speaker_friend,
                    """It's good to see you again.  I'm so sure we can work
                    this thing out together.
                    """)
                self.play(self.sound_friend_friend)

    def after_dialogue(self):
        if self.hand is self.hand_phone:
            self.hand = None
            self.get_prop('phone_up').hidden = True
            self.get_prop('phone_down').hidden = False

        if self.have_boyfriend_offered_keys and not self.have_keys:
            self.have_keys = True
            self.get_prop('keys').hidden = False
            self.play(self.sound_bf_keys)

        if self.have_phoned_mum and not self.have_keys:
            self.set_bar_message(self.speaker_boyfriend,
                """Come on then, let's go to your mum's and sort this
                   thing out once and for all.""")
            self.have_boyfriend_offered_keys = True

        if self.have_phoned_friend and not self.have_friend:
            friend = self.get_prop('friend')
            friend.hidden = False
            self.friend_time = 4.
            def update_friend(dt):
                self.friend_time -= dt
                f = 1 - self.friend_time / 4
                friend.x = (1 - f) * friend.save_x + f * -114
                friend.y = (1 - f) * friend.save_y + f * 89
                friend.z = (1 - f) * friend.save_z + f * -8.4
                friend.scale = (1 - f) * friend.save_scale + f * 1.83
                friend.recreate()
                self.resort_props()
                if self.friend_time <= 0:
                    friendj = self.get_prop('friend_jacket')
                    friendj.x = friend.x
                    friendj.y = friend.y
                    friendj.z = friend.z
                    friendj.scale = friend.scale
                    friendj.recreate()
                    self.resort_props()
                    pyglet.clock.unschedule(update_friend)

            pyglet.clock.schedule(update_friend)
            self.have_friend = True

    def after_panic(self):
        if self.have_boyfriend and self.have_keys:
            self.set_bar_message(self.speaker_boyfriend,
                """Shit, would you stop that?  Didn't you take your
                   medication this morning?""")
        elif self.have_friend:
            self.set_bar_message(self.speaker_friend,
                """Oh wow, you freaked me out there for little while!""")

    def call_mum(self):
        if not self.have_phoned_mum:
            self.set_bar_message(self.speaker_mum,
                  """Darling, I don't want to hear any more of your lies.  Why
                  don't you come over and we can sit down and talk things
                  over?""")
            self.have_phoned_mum = True
            self.play(self.sound_mum)
        else:
            self.set_bar_message(self.speaker_mum,
                  """My goodness, dear!  Are you still at home?  I've been
                  waiting for you.""")
            self.play(self.sound_mum_short)

    def call_friend(self):
        self.set_bar_message(self.speaker_friend,
            """Don't worry, we can figure this thing out.  I'll come over
            right away.""")
        self.have_phoned_friend = True
        self.play(self.sound_friend)

    def rattle_cupboard(self):
        door = self.get_prop('door_c')
        self.door_timeout = 0.5
        x = door.x
        def rattle(dt):
            self.door_timeout -= dt
            door.x = x + math.sin(self.door_timeout * 15) ** 2 * 3
            door.recreate()
            if self.door_timeout < 0:
                door.x = x
                pyglet.clock.unschedule(rattle)
        pyglet.clock.schedule(rattle)
        self.set_bar_message(self.speaker_boyfriend,
            """I thought you took your medication this morning?  Jesus,
            get your act together.
            """)

    def unlock_cupboard(self):
        self.have_unlocked_cupboard = True
        self.get_prop('door_o').hidden = False
        self.get_prop('door_c').hidden = True

    def flush_pills(self):
        pyglet.clock.schedule(self.update_boyfriend)
        self.have_boyfriend = False
        
    def update_boyfriend(self, dt):
        bf = self.get_prop('boyfriend')
        bf.opacity -= 5
        if bf.opacity < 0:
            bf.hidden = True
            pyglet.clock.unschedule(self.update_boyfriend)
        else:
            bf.recreate()

    def is_success(self):
        if self.CHEAT_SUCCESS:
            return True
        return self.hand is self.hand_kite and self.have_warm_friend 

    def go_outside(self):
        friend = self.get_prop('friend')
        friend_save_x = friend.x
        def move_friend(dt):
            friend.x -= dt * 200
            friend.recreate()
            if friend.x < friend_save_x - 200 * 5:
                friend.x = friend_save_x
                friend.recreate()
                pyglet.clock.unschedule(move_friend)

        def move4():
            self.panic_time = 14.
            CameraMotion(self.camera, 
                         self.camera.clone(), 
                         self.cameras[6], 4, move5)
            pyglet.clock.schedule(self.update_panic)
            pyglet.clock.schedule_once(move4_1, 2)

        def move4_1(dt):
            if self.have_friend and not self.is_success():
                pyglet.clock.schedule(move_friend)

        def move5():
            CameraMotion(self.camera, 
                         self.camera.clone(), 
                         self.cameras[8], 4, move6)

        def move6():
            CameraMotion(self.camera, 
                         self.camera.clone(), 
                         self.cameras[9], 6, self.reset_panic)
        self.cutscene = True
        move4()

        if self.is_success():
            self.play(self.sound_end)
        else:
            self.play(self.sound_panic)

    def update_panic(self, dt):
        self.panic_time -= dt

        street = self.get_prop('street')
        if self.is_success():
            f = min(self.panic_time, 4) / 4
            f = math.sin(f * math.pi / 2) ** 2
            street.y = street.save_y - 2600 * (1 - f)
            self.hand_y = (1-f) * -self.hand.height
            self.bg_color = self.BG_COLOR * f
        else:
            self.fade = max(1 - self.panic_time, 0)
            if self.panic_time < 5:
                street.x += (math.sin(self.panic_time * 20) ** 2 - .5) * 6
                street.y += (math.sin(self.panic_time * 4) ** 2 - .5) * 5
            if self.panic_time < 3:
                street.scale = street.save_scale + (3 - self.panic_time)
        street.recreate()

    def reset_panic(self):
        pyglet.clock.unschedule(self.update_panic)
        self.set_camera(0, True)
        self.fade = 1

        street = self.get_prop('street')
        street.scale = street.save_scale
        street.recreate()

        if self.is_success():
            pyglet.clock.schedule_once(self.end_game, 0.5)
        else:
            pyglet.clock.schedule_once(self.begin, 2.)

    def fade_down(self, callback):
        def func(dt):
            self.fade += dt
            if self.fade >= 1:
                self.fade = 1
                pyglet.clock.unschedule(func)
                callback()
        self.fade = 0
        pyglet.clock.schedule(func) 

    def fade_up(self, callback):
        def func(dt):
            self.fade -= dt
            if self.fade < 0:
                self.fade = 0
                pyglet.clock.unschedule(func)
                callback()
        self.fade = 1
        pyglet.clock.schedule(func) 

    def restart_game(self):
        def a():
            self.init_gameplay()
            self.card = self.card_start
            self.bg_color = self.BG_COLOR
            self.labels = []
            pyglet.clock.schedule_once(b, 1.)
        def b(dt):
            self.fade_up(c)
            self.play(self.sound_intro)
        def c():
            self.cutscene = False
        self.cutscene = True
        self.fade_down(a)
        self.fade_down_labels()

    def start_game(self):
        self.cutscene = True
        def f():
            self.card = None
            pyglet.clock.schedule_once(self.first_begin, 1.)
        self.fade_down(f)

    def first_begin(self, dt=None):
        self.cutscene_labels = [
            pyglet.text.Label("I've got to get out of here.",
                              font_name='Imperator',
                              font_size=13,
                              anchor_x='center',
                              x=300, y=200,
                              color=(255,255,255,0)),
        ]
        self.fade_up_labels()
        pyglet.clock.schedule_once(self.fade_down_labels, 3)
        pyglet.clock.schedule_once(self.begin, 5)

    def begin(self, dt=None):
        def f():
            self.cutscene = False
            self.after_panic()
        self.fade_up(f)
        self.play(self.sound_begin)

    def end_game(self, dt=None):
        self.card = self.card_end
        self.card_y = -self.card.height
        self.pan_time = 0.
        def pan_down(dt):
            self.pan_time += dt / 3
            f = math.sin(self.pan_time * math.pi / 2) ** 2
            self.card_y = -(1 - f) * self.card.height
            if self.pan_time >= 1:
                self.card_y = 0
                pyglet.clock.unschedule(pan_down)
                pyglet.clock.schedule_once(self.show_credits, 1)
        self.fade = 0
        pyglet.clock.schedule(pan_down)

    def show_credits(self, dt=None):
        self.cutscene_labels = [
            pyglet.text.Label('Code & music by Alex Holkner',
                              font_name='Imperator',
                              font_size=13,
                              anchor_x='right',
                              x=590, y=40,
                              color=(255,255,255,0)),
            pyglet.text.Label('Art by Rebecca Wong',
                              font_name='Imperator',
                              font_size=13,
                              anchor_x='right',
                              x=590, y=15,
                              color=(255,255,255,0)),
        ]

        pyglet.clock.schedule_once(self.fade_up_labels, 1)
        pyglet.clock.schedule_once(self.fade_down_labels, 4)
        pyglet.clock.schedule_once(self.show_credits2, 5)

    def show_credits2(self, dt=None):
        self.cutscene_labels = [
        pyglet.text.Label('Leaving Home',
                          font_name='Imperator',
                          font_size=13,
                          anchor_x='right',
                          x=590, y=40,
                          color=(255,255,255,0)),
        pyglet.text.Label('PyWeek April 2009',
                          font_name='Imperator',
                          font_size=13,
                          anchor_x='right',
                          x=590, y=15,
                          color=(255,255,255,0)),
        ]

        self.cutscene = False
        pyglet.clock.schedule_once(self.fade_up_labels, 1)

    def fade_up_labels(self, dt=None):
        def f(dt):
            for label in self.cutscene_labels:
                r, g, b, a = label.color
                a += int(dt * 255)
                if a >= 255:
                    pyglet.clock.unschedule(f)
                    a = 255
                label.color = (r, g, b, a)
        pyglet.clock.schedule(f)

    def fade_down_labels(self, dt=None):
        def f(dt):
            for label in self.cutscene_labels:
                r, g, b, a = label.color
                a -= int(dt * 255)
                if a <= 0:
                    pyglet.clock.unschedule(f)
                    a = 0
                label.color = (r, g, b, a)
        pyglet.clock.schedule(f)

    # Dialogue bar
    # ------------

    def init_bar(self):
        self.bar = None
        self.bar_label = pyglet.text.Label(x=150,
                                           width=380,
                                           multiline=True,
                                           font_name='Imperator',
                                           font_size=13,
                                           anchor_y='top',
                                           color=(255, 255, 255, 255))
        self.bar_label.set_style('leading', 10)
        self.bar_speaker = None
        self.next_bar_speaker = None
        self.next_bar_message = None


    def set_bar_message(self, speaker, message):
        message = re.sub(r'[\r\n ]+', ' ', message)

        self.next_bar_speaker = speaker
        self.next_bar_message = message

        if self.bar is not None:
            self.hide_bar()
        else:
            self.show_next_bar_message()

    def show_next_bar_message(self):
        if self.next_bar_message:
            self.bar_label.text = self.next_bar_message
            self.bar_speaker = self.next_bar_speaker
            self.next_bar_message = None
            self.show_bar()

    def show_bar(self):
        if self.bar is not None:
            return

        self.bar = self.bar_start_y = -self.bar_image.height - 20
        self.bar_dest_y = max(
            self.bar_label.content_height + 40,
            self.bar_speaker.height - 10) - self.bar_image.height
        self.bar_f = 0.
        pyglet.clock.schedule(self.update_bar)
        pyglet.clock.unschedule(self.timeout_bar)
        pyglet.clock.schedule_once(self.timeout_bar, 
            .05 * len(self.bar_label.text))

    def hide_bar(self):
        if self.bar is None or self.bar_dest_y == -self.bar_image.height - 20:
            return

        self.bar_f = 0.
        self.bar_dest_y = -self.bar_image.height - 20
        self.bar_start_y = self.bar
        pyglet.clock.unschedule(self.update_bar)
        pyglet.clock.schedule(self.update_bar)
        pyglet.clock.unschedule(self.timeout_bar)
    
    def update_bar(self, dt):
        self.bar_f += dt
        f = math.sin(self.bar_f * math.pi / 2) ** 2
        self.bar = self.bar_start_y + (self.bar_dest_y - self.bar_start_y) * f
        self.bar_label.y = self.bar + self.bar_image.height - 20
        if self.bar_f >= 1:
            self.bar = self.bar_dest_y
            pyglet.clock.unschedule(self.update_bar)
            if self.bar_dest_y < self.bar_start_y:
                self.bar = None
                self.after_dialogue()
                self.show_next_bar_message()
        self.invalid = True

    def timeout_bar(self, dt):
        self.hide_bar()

    # Scene
    # -----

    def load(self):
        self.props = []
        for line in pyglet.resource.file('scene.txt'):
            filename, mask_filename, x, y, z, scale = line.strip().split()
            x, y, z, scale = map(float, (x, y, z, scale))
            prop = Prop(filename, mask_filename, x, y, z, scale)
            self.props.append(prop)
        self.resort_props()

        self.cameras = []
        for line in pyglet.resource.file('cameras.txt'):
            x, y, z, yaw = line.strip().split()
            x, y, z, yaw = map(float, (x, y, z, yaw))
            camera = Camera(x, y, z, yaw)
            self.cameras.append(camera)

        self.border = pyglet.resource.image('border.png')
        self.bar_image = pyglet.resource.image('bar.png')

        self.speaker_boyfriend = pyglet.resource.image('head_bf.png')
        self.speaker_friend = pyglet.resource.image('head_friend.png')
        self.speaker_mum = pyglet.resource.image('head_mum.png')

        self.hand_pills = pyglet.resource.image('pillshere.png')
        self.hand_phone = pyglet.resource.image('phonehere.png')
        self.hand_yearbook = pyglet.resource.image('yearbook.png')
        self.hand_jacket = pyglet.resource.image('jackethere.png')
        self.hand_kite = pyglet.resource.image('kitehere.png')
        self.hand_keys = pyglet.resource.image('keyshere.png')
        self.hand_yearbook.anchor_x = self.hand_yearbook.width - self.width

        self.card_end = pyglet.resource.image('end.png')
        self.card_start = pyglet.resource.image('menu.png')

    def resort_props(self):
        self.props.sort(key=lambda p: p.z)

    def get_prop(self, name):
        filename = name + '.png'
        return [prop for prop in self.props if prop.filename == filename][0]

    def save(self):
        f = pyglet.resource.file('scene.txt', 'w')
        for prop in self.props:
            f.write('%s %s %f %f %f %f\n' % 
                (prop.filename, prop.mask_filename, 
                 prop.x, prop.y, prop.z, prop.scale))
        f = pyglet.resource.file('cameras.txt', 'w')
        for camera in self.cameras:
            f.write('%f %f %f %f\n' % 
                (camera.x, camera.y, camera.z, camera.yaw))
        print 'Saved.'

    def init_gl(self):
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    # Drawing
    # -------

    def set_camera(self, index, immediate=False):
        if immediate:
            self.camera.set_from(self.cameras[index])
        else:
            CameraMotion(self.camera, self.camera.clone(), self.cameras[index])

    def on_draw(self):
        self.invalid = False

        # Loading screen
        if self.loading:
            pyglet.clock.schedule(self.load_all)

        # Selection

        if self.mouse_press:
            self.apply_camera()

            glEnable(GL_ALPHA_TEST)
            glAlphaFunc(GL_GREATER, 0.9)
            glClearColor(1, 0, 1, 1)

            # Wow, ok, this is fairly embarrasingly inefficient, but it
            # was easier than convincing Biccy to draw white masks I could
            # color-code :-)
            x, y = self.mouse_press
            selected_prop = None
            for prop in self.props:
                glClear(GL_COLOR_BUFFER_BIT)
                prop.draw_hit_test()
                data = (GLubyte * 4)()
                glFlush()
                glReadPixels(x, y, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, data)
                if tuple(data[:3]) != (255, 0, 255):
                    selected_prop = prop
            if selected_prop:
                if selected_prop.filename in (
                    'grass.png', 'fence_left.png',
                    'fence_right.png', 'street.png'):
                    self.select_outside()
                else:
                    self.select_prop(selected_prop)
            else:
                self.select_outside()

            glDisable(GL_ALPHA_TEST)
            self.mouse_press = None

        glClearColor(self.bg_color, self.bg_color, self.bg_color, 1)
        self.clear()

        if self.loading:
            pass
        elif self.card:
            self.apply_hud()
            self.card.blit(0, self.card_y)
        else:
            # Scene
            self.apply_camera()
            for prop in self.props:
                prop.draw()

            self.apply_hud()

            # Border
            self.border.blit(0, 0)

            # Hand
            if self.hand:
                self.hand.blit(0, self.hand_y)

            # Dialogue bar
            if self.bar is not None:
                self.bar_image.blit(0, self.bar)
                self.bar_speaker.blit(10, 
                    self.bar + self.bar_image.height - 
                    self.bar_speaker.height + 20)
                self.bar_label.draw()

        # Fade to black
        if self.fade:
            glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT)
            glDisable(GL_TEXTURE_2D)
            glColor4f(0, 0, 0, self.fade)
            glRecti(0, 0, self.width, self.height)
            glPopAttrib()

        # Labels
        if self.cutscene_labels:
            self.apply_hud()
            for label in self.cutscene_labels:
                label.draw()

    def apply_camera(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        self.apply_selection_viewport()
        gluPerspective(100., self.width / float(self.height), .1, 1000)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glTranslatef(-self.camera.x, -self.camera.y, -self.camera.z)
        glRotatef(self.camera.yaw, 0, 1, 0)
        glScalef(.01, .01, 1)

    def apply_hud(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, self.width, 0, self.height, -1, 1)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def apply_selection_viewport(self):
        if self.mouse_press:
            x, y = self.mouse_press
            viewport = (GLint * 4)()
            glGetIntegerv(GL_VIEWPORT, viewport)
            gluPickMatrix(x, y, 4, 4, viewport)

    def on_mouse_press(self, x, y, button, modifiers):
        if not self.cutscene:
            if self.card is self.card_start:
                self.start_game()
            elif self.card is self.card_end:
                self.restart_game()
            elif self.bar:
                self.hide_bar()
            else:
                self.mouse_press = x, y
                self.invalid = True

    def on_activate(self):
        self.invalid = True
    
    def on_expose(self):
        self.invalid = True

    # Sound
    # -----

    def init_sound(self):
        if ENABLE_SOUND:
            self.player = pyglet.media.Player()
            self.old_player = pyglet.media.Player()

        def load(name):
            if ENABLE_SOUND:
                return pyglet.resource.media(name, streaming=False)

        self.sound_intro = load('intro.mp3')
        self.sound_begin = load('begin.mp3')
        self.sound_panic = load('panic.mp3')
        self.sound_end = load('end.mp3')
        self.sound_bf_bf = load('bf_bf.mp3')
        self.sound_bf_cupboard = load('bf_cupboard.mp3')
        self.sound_bf_outside = load('bf_outside.mp3')
        self.sound_bf_sink = load('bf_sink.mp3')
        self.sound_bf_mum = load('bf_mum.mp3')
        self.sound_bf_keys = load('bf_keys.mp3')
        self.sound_pickup = load('pickup.mp3')
        self.sound_putdown = load('putdown.mp3')
        self.sound_mum = load('mum.mp3')
        self.sound_mum_short = load('mum_short.mp3')
        self.sound_unlock = load('unlock.mp3')
        self.sound_flush = load('flush.mp3')
        self.sound_friend = load('friend.mp3')
        self.sound_friend_mum = load('mum_short.mp3')
        self.sound_friend_friend = load('friend_friend.mp3')
        self.sound_friend_sink = load('friend_sink.mp3')
        self.sound_jacket = load('jacket.mp3')

    def play(self, sound, interrupt=True):
        if not ENABLE_SOUND:
            return

        # If a sound is currently playing, fade it out quickly while playing
        # the new sound immediately
        def fade_out(dt):
            volume = self.old_player.volume - dt * 2
            if volume <= 0:
                self.old_player.next()
                self.old_player.volume = 1.
                pyglet.clock.unschedule(fade_out)
            else:
                self.old_player.volume = volume

        if self.player.playing:
            self.player, self.old_player = self.old_player, self.player
            pyglet.clock.schedule(fade_out)

        self.player.next()
        self.player.queue(sound)
        self.player.play()

class GameDebug(object):
    def __init__(self, game):
        self.game = game
        self.selected_prop = None
        self.scheduled_func = None
        self.precision = 0.02

    def change_selection(self, direction):
        index = 0
        if self.selected_prop in self.game.props:
            index = self.game.props.index(self.selected_prop)

        index += direction
        self.select_prop(self.game.props[index % len(self.game.props)])

    def select_prop(self, prop):
        print prop.filename
        if self.selected_prop:
            self.selected_prop.opacity = 255
            self.selected_prop.recreate()
        self.selected_prop = prop
        if self.selected_prop:
            self.selected_prop.opacity = 127
            self.selected_prop.recreate()

    def move_prop(self, dx, dy, dz):
        prop = self.selected_prop
        if not prop:
            return

        def move(dt):
            prop.x += dx * self.precision
            prop.y += dy * self.precision
            prop.z += dz * self.precision
            prop.recreate()
            if dz:
                game.resort_props()
        move(0)
        self.schedule(move)

    def scale_prop(self, scale):
        prop = self.selected_prop
        if not prop:
            return

        def move(dt):
            prop.scale += scale * self.precision
            prop.recreate()
        move(0)
        self.schedule(move)

    def move_camera(self, dx, dy, dz, dyaw):
        def move(dt):
            self.game.camera.x += dx * self.precision
            self.game.camera.y += dy * self.precision
            self.game.camera.z += dz * self.precision
            self.game.camera.yaw += dyaw * self.precision

        move(0)
        self.schedule(move)

    def save_camera(self, index):
        self.game.cameras[index].set_from(self.game.camera)
        print 'Set camera', index

    def schedule(self, func):
        self.unschedule()
        self.scheduled_func = func
        pyglet.clock.schedule(func)

    def unschedule(self):
        if self.scheduled_func:
            pyglet.clock.unschedule(self.scheduled_func)
            self.scheduled_func = None

    def on_key_press(self, symbol, modifiers):
        if symbol == key.LCTRL:
            self.precision = 10.0
            return

        if modifiers & key.MOD_ACCEL:
            if symbol >= key._0 and symbol <= key._9:
                if modifiers & key.MOD_SHIFT:
                    self.save_camera(symbol - key._0)
                else:
                    self.game.set_camera(symbol - key._0)
            elif symbol == key.S:
                self.game.save()

        elif modifiers & key.MOD_SHIFT:
            if symbol == key.LEFT:
                self.move_camera(-1, 0, 0, 0)
            elif symbol == key.RIGHT:
                self.move_camera(1, 0, 0, 0)
            elif symbol == key.UP:
                self.move_camera(0, 1, 0, 0)
            elif symbol == key.DOWN:
                self.move_camera(0, -1, 0, 0)
            elif symbol == key.W:
                self.move_camera(0, 0, -.1, 0)
            elif symbol == key.S:
                self.move_camera(0, 0, .1, 0)
            elif symbol == key.D:
                self.move_camera(0, 0, 0, 1)
            elif symbol == key.A:
                self.move_camera(0, 0, 0, -1)
            elif symbol == key.X:
                self.game.camera_yaw = 0
        else:
            if symbol == key.PERIOD:
                self.change_selection(1)
            elif symbol == key.COMMA:
                self.change_selection(-1)
            elif symbol == key.LEFT:
                self.move_prop(-1, 0, 0)
            elif symbol == key.RIGHT:
                self.move_prop(1, 0, 0)
            elif symbol == key.UP:
                self.move_prop(0, 1, 0)
            elif symbol == key.DOWN:
                self.move_prop(0, -1, 0)
            elif symbol == key.W:
                self.move_prop(0, 0, -.2)
            elif symbol == key.S:
                self.move_prop(0, 0, .2)
            elif symbol == key.EQUAL:
                self.scale_prop(0.1)
            elif symbol == key.MINUS:
                self.scale_prop(-0.1)
            elif symbol == key.ESCAPE:
                self.select_prop(None)

            elif symbol == key.P:
                self.game.CHEAT_SUCCESS = not self.game.CHEAT_SUCCESS
                print 'success', self.game.CHEAT_SUCCESS
            elif symbol == key.L:
                self.game.call_friend()
            elif symbol == key.O:
                self.game.flush_pills()

        self.game.invalid = True
        return True

    def on_key_release(self, symbol, modifiers):
        if symbol == key.LCTRL:
            self.precision = 0.02
        else:
            self.unschedule()

if __name__ == '__main__':
    try:
        import pyglet.media.avbin
        pyglet.media.avbin.get_version()
    except:
        msg = ('There was a problem loading AVbin, which is needed to play ' +
            'sounds.  Would you like to play the game anyway, without sound?')
        response = embryo.message_box(msg, 'Leaving Home', cancel_button=True)
        if not response:
            sys.exit(0)
        else:
            ENABLE_SOUND = False

    try:
        game = Game(caption='Leaving Home', 
                    width=600,
                    height=400)
        pyglet.app.run()
    except:
        embryo.message_box('Oops... an error occurred.  The error was:\n\n' +
            traceback.format_exc(limit=1), 'Leaving Home Fail')
        print >> open('error.txt', 'wt'), traceback.format_exc()
